@@ -15,6 +15,10 @@ class AlreadyDefinedEntityError(ValueError):
         return cls(f"Entity {name!r} is already defined")
 
 
+class UnsetEntityError(ValueError):
+    pass
+
+
 class IncompatibleEntityError(ValueError):
     pass
 
@@ -37,3 +41,7 @@ class EntityComputationError(Exception):
 
 class AttributeValidationError(Exception):
     pass
+
+
+class MalformedDescriptorError(Exception):
+    pass
