@@ -28,7 +28,7 @@ from .descriptors.parsing import entity_dnode_from_descriptor
 from .descriptors import ast
 from .bytecode import canonical_bytecode_bytes_from_func
 from .util import groups_dict, hash_to_hex, oneline
-from .optdep import import_optional_dependency
+from .deps.optdep import import_optional_dependency
 
 import logging
 
@@ -37,18 +37,12 @@ logger = logging.getLogger(__name__)
 
 class ProviderAttributes:
     def __init__(
-        self,
-        dnodes,
-        code_version=None,
-        orig_flow_name=None,
-        is_default_value=None,
-        changes_per_run=None,
+        self, dnodes, code_version=None, orig_flow_name=None, changes_per_run=None,
     ):
         self.dnodes = dnodes
         self.code_version = code_version
         self.orig_flow_name = orig_flow_name
         self.changes_per_run = changes_per_run
-        self.is_default_value = is_default_value
 
 
 class BaseProvider:
